#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <dirent.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include "iolib.h"
const char *progname = "svscan";
#include "freedt.h"

struct service {
	char *dir;
	int logpipe[2];
	int found;
	struct service *next;
};

struct service *services = NULL;
int numservices = 0;
int startdir;

void sigchld_handler(int value) {
	while (waitpid(0, NULL, WNOHANG) > 0)
		/* nothing */;
}

int service_running(char *dir) {
	int fd, dirfd, rc = 1;

	dirfd = open(".", O_RDONLY);
	if (dirfd < 0) {
		warn2(dir, "unable to open service parent directory");
		return -1;
	}

	if (chdir(dir) < 0) {
		warn2(dir, "unable to change into service directory");
		return -1;
	}

	fd = open("supervise/control", O_WRONLY | O_NONBLOCK);
	if (fd < 0) {
		if (errno == ENXIO || errno == ENOENT) {
			rc = 0;
		} else {
			warn2(dir, "unable to open control pipe");
			return -1;
		}
	} else {
		rc = 1;
		close(fd);
	}


	if (fchdir(dirfd) < 0) {
		warn2(dir, "unable to change back to parent directory");
		close(dirfd);
		return -1;
	}
	close(dirfd);

	return rc;
}

void check_supervise(char *dir, int in, int err) {
	pid_t pid;

	if (service_running(dir))
		return;

	pid = fork();
	if (pid < 0) {
		warn2(dir, "fork failed");
		return;
	}

	if (pid == 0) {
		struct service *s;

		if (in != -1) {
			if (dup2(in, STDIN_FILENO) < 0)
				_exit(111);
		}
		if (err != -1) {
			if (dup2(err, STDOUT_FILENO) < 0)
				_exit(111);
		}
		close(startdir);

		for (s = services; s != NULL; s = s->next) {
			if (s->logpipe[0] != -1) {
				close(s->logpipe[0]);
				close(s->logpipe[1]);
			}
		}

		execlp("supervise", "supervise", dir, NULL);
		_exit(111);
	}
}

void check_service(struct service *serv) {
	if (fchdir(startdir) < 0) {
		warn2(serv->dir, "unable to change back to"
			"specified directory");
		return;
	}
	check_supervise(serv->dir, -1, serv->logpipe[1]);
	if (serv->logpipe[0] != -1) {
		if (chdir(serv->dir) < 0) {
			warn2(serv->dir, "unable to change into"
				"service directory");
			return;
		}
		check_supervise("log", serv->logpipe[0], -1);
	}
}

struct service *add_service(char *dir) {
	struct stat st;
	struct service *svc;

	if (fchdir(startdir) < 0) {
		warn2(dir, "unable to change back to specified directory");
		return NULL;
	}

	if (stat(dir, &st) < 0) {
		warn2(dir, "unable to stat dir");
		return NULL;
	}

	svc = malloc(sizeof *svc);
	if (svc == NULL) {
		warn("out of memory");
		return NULL;
	}

	if (st.st_mode & S_ISVTX) {
		if (pipe(svc->logpipe) < 0) {
			warn("cannot create pipe");
			free(svc);
			return NULL;
		}
	} else {
		svc->logpipe[0] = -1;
		svc->logpipe[1] = -1;
	}


	svc->dir = strdup(dir);
	svc->next = NULL;
	if (services == NULL) {
		services = svc;
	} else {
		struct service *last;
		for (last = services; last->next; last = last->next)
			;
		last->next = svc;
	}

	check_service(svc);
	return svc;
}

void remove_service(struct service *this) {
	struct service *prev;

	if (services == this) {
		services = this->next;
	} else {
		for (prev = services; prev; prev = prev->next)
			if (prev->next == this)
				break;
		prev->next = this->next;
	}
	if (this->logpipe[0] != -1) {
		close(this->logpipe[0]);
		close(this->logpipe[1]);
	}
	free(this->dir);
	free(this);
}

struct service *find_service(char *dir) {
	struct service *svc;

	for (svc = services; svc; svc = svc->next) {
		if (strcmp(svc->dir, dir) == 0)
			return svc;
	}
	return NULL;
}

void reliable_sleep(int seconds) {
	while (seconds > 0)
		seconds = sleep(seconds);
}

int main(int argc, char **argv) {
	struct sigaction sa;
	DIR *dir;

	istdinit();

	if (argc > 1) {
		if (chdir(argv[1]) < 0)
			die("unable to chdir to specified directory");
	}

	startdir = open(".", O_RDONLY);
	if (startdir < 0)
		die("unable to open specified directory");

	sa.sa_handler = sigchld_handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	if (sigaction(SIGCHLD, &sa, NULL) < 0)
		die("unable to install SIGCHLD handler");

	while (1) {
		struct service *svc;

		if (fchdir(startdir) < 0) {
			warn("unable to change back to specified directory");
			goto next;
		}

		dir = opendir(".");
		if (dir == NULL) {
			warn("unable to list directory");
			goto next;
		}

		for (svc = services; svc; svc = svc->next)
			svc->found = 0;

		while (1) {
			struct dirent *entry = readdir(dir);

			if (entry == NULL)
				break;
			if (entry->d_name[0] == '.')
				continue;

			svc = find_service(entry->d_name);
			if (svc)
				check_service(svc);
			else
				svc = add_service(entry->d_name);
			if (svc)
				svc->found = 1;
		}

		for (svc = services; svc;) {
			struct service *next = svc->next;
			if (svc->found == 0)
				remove_service(svc);
			svc = next;
		}

		closedir(dir);
	next:
		reliable_sleep(5);
	}

	return 0; /* NOTREACHED */
}

